import os

from pytest import fixture
from staze.core.assembler.assembler import Assembler
from staze.core.assembler.build import Build
from staze.core.app.app_mode_enum import RunAppModeEnum
from staze.core.app.app import App
from staze.core.database.database import Database
from staze.core.log import log
from staze.core.test.test import Test
from staze.tests.blog.app.user.user_service import UserService


@fixture
def assembler_test(
        blog_build: Build, default_host: str, default_port: int,
        blog_root_dir: str):
    assembler = Assembler(
        build=blog_build,
        mode_enum=RunAppModeEnum.TEST,
        host=default_host,
        port=default_port,
        root_dir=os.path.join(os.getcwd(), blog_root_dir))

    yield assembler

    assembler.cleanup_all_services()
    assembler.__class__.instances = {}


@fixture
def assembler_dev(
        blog_build: Build, default_host: str, default_port: int,
        blog_root_dir: str):
    assembler = Assembler(
        build=blog_build,
        mode_enum=RunAppModeEnum.DEV,
        host=default_host,
        port=default_port,
        root_dir=os.path.join(os.getcwd(), blog_root_dir))

    yield assembler

    assembler.cleanup_all_services()
    assembler.__class__.instances = {}


@fixture
def assembler_prod(
        blog_build: Build, default_host: str, default_port: int,
        blog_root_dir: str):
    assembler = Assembler(
        build=blog_build,
        mode_enum=RunAppModeEnum.PROD,
        host=default_host,
        port=default_port,
        root_dir=os.path.join(os.getcwd(), blog_root_dir))

    yield assembler

    assembler.cleanup_all_services()
    assembler.__class__.instances = {}

    Assembler.instances = {}


class TestAssembler(Test):
    def test_create_manual(self,
            blog_build: Build, default_host: str, default_port: int,
            blog_root_dir: str):
        assembler = Assembler(
            build=blog_build,
            mode_enum=RunAppModeEnum.TEST,
            host=default_host,
            port=default_port,
            root_dir=os.path.join(os.getcwd(), blog_root_dir))
        
        assert assembler.mode_enum == RunAppModeEnum.TEST

    def test_build(self, assembler_dev: Assembler):
        app: App = App.instance()
        database: Database = Database.instance()
        user_service: UserService = assembler_dev.custom_services['user']



from dataclasses import dataclass, fields
from typing import Any, Iterable
import types
from staze.core.log import log
from typing import ClassVar
from warepy import snakefy
from pydantic import BaseModel


class Model(BaseModel):
    """Basic interface dataclass."""
    # Name to be excluded from final formatted name
    _base_name: str = 'Model'
    _formatted_name: str | None = None

    @property
    def formatted_name(self) -> str:
        return self._get_formatted_name()

    @property
    def api_dict(self) -> dict:
        return self.get_api_dict()

    def get_api_dict(
            self,
            formatted_name: str | None = None,
            is_without_formatted_name: bool = False) -> dict:
        """Translates interface to dict structure for API.
        
        Args:
            formatted_name (str, optional):
                Custom formatted name as external dictionary key.
                Defaults to class's base defined (property formatted_name).
            is_without_formatted_name(str, optional):
                If True, dict without external key is returned, but note that
                nested models still will have such external key.
                If you don't need any external keys for nested models, use
                standard Pydantic dict() instead.
        Returns:
            dict:
                Dictionary with interface decomposed to structure for API:
        ```
        {
           model_formatted_name: {
                ...model structure    
            }
        }
        ```
        """
        _formatted_name: str
        formatted_dict: dict = {}

        if formatted_name:
            _formatted_name = formatted_name  
        else:
            _formatted_name = self._get_formatted_name()

        for k, v in self.__dict__.items():
            alias: str = k
            if isinstance(v, Model):
                alias = v.formatted_name
            
            formatted_dict[alias] = self._parse_value(v)

        if is_without_formatted_name:
            return formatted_dict
        else:
            return {
                _formatted_name: formatted_dict
            }

    def _parse_value(
            self, v: Any) -> Any:
        final_value: Any

        if isinstance(v, Model):
            final_value = v.get_api_dict(is_without_formatted_name=True)
        # Not passed for cases of missing attr and empty dict.
        # I.e. classes with empty dict returned from __dict__ skipping
        # this branch
        elif getattr(v, "__dict__", None):
            final_value = v.__dict__
        elif \
                isinstance(v, Iterable) \
                and (hasattr(v, "append") or hasattr(v, "push")):
            final_value = type(v)()
            for element in v:
                parsed_element: Any
                if isinstance(element, Model):
                    parsed_element = {
                        element.formatted_name: self._parse_value(element)
                    }
                else:
                    parsed_element = self._parse_value(element)

                if hasattr(v, "append"):
                    # Call again but for every list element
                    final_value.append(parsed_element)
                else:
                    # In theory, set could be iterable with `push` operation
                    # but set cannot store unhashable types so it doesn't
                    # make sense. But custom iterables may still implement
                    # `push` method
                    final_value.push(parsed_element)
        else:
            final_value = v

        return final_value

    @classmethod
    def _get_formatted_name(cls) -> str:
        name: str

        if cls._formatted_name:
            name = cls._formatted_name
        else:
            name = snakefy(cls.__name__.replace(cls._base_name, ''))

        return name



import sys
from io import StringIO 

from pytest import fixture
from staze.core.app.app_mode_enum import RunAppModeEnum
from staze.core.test.test import Test
from staze.core.assembler.build import Build
from staze.core.validation import validate_re
from staze.core.cli.cli import Cli
from staze.core.log import log
from staze.core.assembler.assembler import Assembler


# FIXME: Remove and import from warepy as it is introduced there
class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout


@fixture
def cli() -> Cli:
    return Cli()


@fixture
def cli_blog(blog_root_dir: str, blog_build: Build):
    yield Cli(root_dir=blog_root_dir, build=blog_build)
    assembler: Assembler = Assembler.instance()
    assembler.cleanup_all_services()
    assembler.__class__.instances = {}


class TestCliExecute(Test):
    def test_default(self, cli: Cli):
        with Capturing() as out:
            try:
                cli.execute(['staze', 'version'])
            except SystemExit:
                pass
            else:
                raise AssertionError('')
        assert len(out) == 1
        validate_re(out[0], r'Staze \d+\.\d+\.\d+')

    def test_test(self, cli_blog: Cli):
        assembler: Assembler = cli_blog.execute(
            ['staze', 'test'], has_to_run_app=False)
        assert assembler.mode_enum.value == 'test'

    def test_dev(self, cli_blog: Cli):
        assembler: Assembler = cli_blog.execute(
            ['staze', 'dev'], has_to_run_app=False)
        assert assembler.mode_enum.value == 'dev'

    def test_prod(self, cli_blog: Cli):
        assembler: Assembler = cli_blog.execute(
            ['staze', 'prod'], has_to_run_app=False)
        assert assembler.mode_enum.value == 'prod'

    def test_host_port(self, cli_blog: Cli):
        assembler: Assembler = cli_blog.execute(
            ['staze', 'dev', '-h', '0.0.0.0', '-p', '6000'], has_to_run_app=False)
        # FIXME: ...
        # assert assembler.mode_enum.value == 'dev'
        # assert assembler.app.host == '0.0.0.0'
        # assert assembler.app.port == 6000
        # assert False



import os
import re
import sys
import pytest
from typing import get_args

from warepy import (
    match_enum_containing_value,
    get_enum_values, get_union_enum_values
)
from staze.core.assembler.build import Build
from staze.core.cli.cli_error import CLIError
from staze.core.log import log

from staze import __version__ as staze_version
from staze.core.assembler.assembler import Assembler
from staze.core import validation, parsing
from staze.core.app.app_mode_enum import (
    RunAppModeEnum, DatabaseAppModeEnum, HelperAppModeEnum, AppModeEnumUnion)
from staze.core.cli.cli_input import CliInput


class Cli():
    def __init__(
                self,
                build: Build | None = None,
                root_dir: str | None = None,
            ) -> None:
        self.build = build
        self.root_dir = root_dir

    def execute(
                self,
                args: list[str] | None = None,
                has_to_run_app: bool = True
            ) -> Assembler:
        cli_input: CliInput = self._parse_input(args)

        match cli_input.mode_enum:
            case HelperAppModeEnum.VERSION:
                print(f"Staze {staze_version}")
                exit()
            case _:
                root_dir: str
                if self.root_dir:
                    root_dir = self.root_dir
                else:
                    root_dir = os.getcwd()

                # Create and build assembler
                assembler = Assembler(
                    mode_enum=cli_input.mode_enum,
                    mode_args=cli_input.mode_args,
                    host=cli_input.host,
                    port=cli_input.port,
                    root_dir=root_dir,
                    build=self.build)

                if has_to_run_app:
                    assembler.run()

                return assembler

    def _parse_input(self, args: list[str] | None = None) -> CliInput:
        if not args:
            args = sys.argv

        check_other_args: bool = True
        mode_kwargs: dict = {}
        mode_kwargs['mode_args'] = []

        match args[1]:
            case 'version':
                if len(args) > 2:
                    raise CLIError(
                        'Mode `version` shouldn\'t be'
                        ' followed by any other arguments')

                mode_enum_class = HelperAppModeEnum
                check_other_args = False 
            case _:
                try:
                    # Find enum where mode assigned
                    mode_enum_class = match_enum_containing_value(
                        args[1], *get_args(AppModeEnumUnion))
                except ValueError:
                    raise CLIError(
                        f'Unrecognized mode: {args[1]}')

        # Create according enum with mode value
        mode_kwargs['mode_enum'] = mode_enum_class(args[1])

        if check_other_args:
            is_previous_flag: bool = False

            for ix, arg in enumerate(args[2:]):
                ix = ix + 2

                match arg:
                    case '-h':
                        if not isinstance(
                                mode_kwargs['mode_enum'], RunAppModeEnum):
                            raise CLIError(
                                'Flag -h applicable only to Run modes:'
                                f' {get_enum_values(RunAppModeEnum)}')
                        elif '-h' in mode_kwargs:
                            raise CLIError('Flag -h has been defined twice')
                        else:
                            try:
                                host = args[ix+1]
                            except KeyError:
                                raise CLIError(
                                    'No host specified for defined flag -h')
                            else:
                                # Pattern from: https://stackoverflow.com/a/36760050
                                validation.validate_re(
                                    host,
                                    r'^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)(\.(?!$)|$)){4}$')
                            mode_kwargs['host'] = host
                            is_previous_flag = True
                    case '-p':
                        if not isinstance(
                                mode_kwargs['mode_enum'], RunAppModeEnum):
                            raise CLIError(
                                'Flag -p applicable only to Run modes:'
                                f' {get_enum_values(RunAppModeEnum)}')
                        elif '-p' in mode_kwargs:
                            raise CLIError('Flag -p has been defined twice')
                        else:
                            try:
                                port = args[ix+1]
                            except KeyError:
                                raise CLIError(
                                    'No port specified for defined flag -p')
                            else:
                                validation.validate_re(
                                    port,
                                    r'^\d+$')
                            mode_kwargs['port'] = port
                            is_previous_flag = True
                    case '-x':
                        # Mode -x applicable to pytest flag and to dev and
                        # prod modes as additional functions executor
                        executable_func_names: list[str] = []

                        if not isinstance(
                                mode_kwargs['mode_enum'], RunAppModeEnum):
                            raise CLIError(
                                'Flag -x can only be used with run modes:'
                                f' {get_enum_values(RunAppModeEnum)}') 

                        # pytest by itself handles -x flag, other cases is up to us
                        if mode_kwargs['mode_enum'] is RunAppModeEnum.TEST:
                            mode_kwargs['mode_args'].append(arg)
                        else:
                            for name in args[ix+1:]:
                                if '-' in name:
                                    break
                                executable_func_names.append(name)

                            if executable_func_names == []:
                                raise CLIError(
                                    'Flag -x for modes prod and dev')

                            mode_kwargs['executable_func_names'].append(
                                executable_func_names)
                            is_previous_flag = True
                    case _:
                        if arg[0] == '-':
                            raise CLIError(f'Unrecognized flag: {arg}')
                        # If value not used by any flag,
                        # e.g. `-h 0.0.0.0 value_without_flag`, raise exception
                        # for run modes
                        elif not is_previous_flag:
                            if mode_enum_class is RunAppModeEnum:
                                if \
                                        mode_kwargs['mode_enum'] \
                                            is not RunAppModeEnum.TEST:
                                    raise CLIError(
                                        'Values without flags is not '
                                        'applicable to mode '
                                        + mode_kwargs['mode_enum'].value
                                    )
                # Write all mode args as it is for extensions read,
                # e.g. for pytest 
                mode_kwargs['mode_args'].append(arg)

        return CliInput(**mode_kwargs)


if __name__ == "__main__":
    Cli().execute()



from pytest import fixture
from staze.core.model.model import Model
from staze.core.test.mock import Mock
from staze.core.log import log


class CustomModel(Model):
    name: str
    age: int
    tags: set[str]


class NestingCustomModel(Model):
    custom_model: CustomModel


class NestingListCustomModel(Model):
    custom_models: list[CustomModel]


class CustomModelMock(Mock):
    name: str
    age: int
    tags: set[str]


@fixture
def custom_model_mock() -> CustomModelMock:
    return CustomModelMock(name='Test', age=12, tags={'hero', 'reckless'})


@fixture
def custom_model(custom_model_mock: CustomModelMock) -> CustomModel:
    return CustomModel(
        name=custom_model_mock.name,
        age=custom_model_mock.age,
        tags=custom_model_mock.tags)


class TestModel():
    def test_custom(
            self, custom_model: CustomModel,
            custom_model_mock: CustomModelMock):
        expected_json: dict = {
            'custom': {
                'name': custom_model_mock.name,
                'age': custom_model_mock.age,
                'tags': custom_model_mock.tags
            }
        }
        log.debug(expected_json)

        assert custom_model.name == custom_model_mock.name
        assert custom_model.age == custom_model_mock.age
        assert custom_model.tags == custom_model_mock.tags

        assert custom_model.formatted_name == 'custom'
        assert custom_model.get_api_dict() == expected_json

        json_custom_name: dict = custom_model.get_api_dict('testname')
        assert 'testname' in json_custom_name

    def test_nesting_custom(
        self, custom_model_mock: CustomModelMock, custom_model: CustomModel
    ):
        expected: dict = {
            'nesting_custom': {
                'custom': {
                    'name': custom_model_mock.name,
                    'age': custom_model_mock.age,
                    'tags': custom_model_mock.tags
                }
            }
        }

        target: NestingCustomModel = NestingCustomModel(
            custom_model=custom_model
        )

        assert target.formatted_name == 'nesting_custom'
        assert target.api_dict == expected
        a = [{'t': 1}, {'b': 2}]

    def test_nesting_list_custom(self):
        expected: dict = {
            'nesting_list_custom': {
                'custom_models': [
                    {
                        'custom': {
                            'name': '1',
                            'age': 1,
                            'tags': {'1'}
                        }
                    },
                    {
                        'custom': {
                            'name': '2',
                            'age': 2,
                            'tags': {'1', '2'}
                        }
                    },
                    {
                        'custom': {
                            'name': '3',
                            'age': 3,
                            'tags': {'1', '2', '3'}
                        }
                    }
                ]
            }
        }

        target = NestingListCustomModel(
            custom_models=[
                CustomModel(name='1', age=1, tags={'1'}),
                CustomModel(name='2', age=2, tags={'1', '2'}),
                CustomModel(name='3', age=3, tags={'1', '2', '3'})
            ]
        )

        assert target.api_dict == expected
